/*
	*** Transformation Commands
	*** src/command/transform.cpp
	Copyright T. Youngs 2007-2015

	This file is part of Aten.

	Aten is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Aten is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Aten.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "command/commands.h"
#include "parser/commandnode.h"
#include "parser/tree.h"
#include "model/model.h"
#include "classes/prefs.h"

// Rotate selection about specified axis / origin
bool Command::function_AxisRotate(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	double angle;
	Atom *i, *j;
	Vec3<double> v, o;
	// Determine which data has been supplied
	switch (c->nArgs())
	{
		// Two atom ids and theta
		case (3):
			i = obj.rs()->atom(c->argi(0)-1);
			j = obj.rs()->atom(c->argi(1)-1);
			if ((i == NULL) || (j == NULL)) return FALSE;
			v = obj.rs()->cell()->mimVector(i,j);
			angle = c->argd(2);
			break;
		// Axis and theta
		case (4):
			v.set(c->argd(0), c->argd(1), c->argd(2));
			angle = c->argd(3);
			break;
		// Two atom ids, theta, and an origin
		case (6):
			i = obj.rs()->atom(c->argi(0)-1);
			j = obj.rs()->atom(c->argi(1)-1);
			if ((i == NULL) || (j == NULL)) return FALSE;
			v = obj.rs()->cell()->mimVector(i,j);
			angle = c->argd(2);
			o.set(c->argd(3), c->argd(4), c->argd(5));
			break;
		// Axis, theta, and origin
		case (7):
			v.set(c->argd(0), c->argd(1), c->argd(2));
			angle = c->argd(3);
			o.set(c->argd(4), c->argd(5), c->argd(6));
			break;
		default:
			msg.print("Odd number of arguments given to 'axisrotate'.\n");
			return FALSE;
			break;
	}
	obj.rs()->beginUndoState("Rotate %i atom(s)", obj.rs()->nSelected());
	obj.rs()->rotateSelectionVector(o, v, angle);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Centre selection at given coordinates
bool Command::function_Centre(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	if (c->parent()->parser() == NULL)
	{
		Vec3<double> centre = c->arg3d(0);
		Vec3<int> lock(0,0,0);
		if (c->hasArg(5)) lock = c->arg3i(3);
		obj.rs()->beginUndoState("Centre %i atom(s) at %f %f %f", obj.rs()->nSelected(), centre.x, centre.y, centre.z);
		obj.rs()->centre(centre, lock.x, lock.y, lock.z);
		obj.rs()->endUndoState();
	}
	else if (prefs.centreOnLoad() != Choice::No) obj.rs()->centre(c->arg3d(0));
	rv.reset();
	return TRUE;
}

// Flip selection's x-coordinates
bool Command::function_FlipX(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	obj.rs()->beginUndoState("Flip x-coordinates of %i atom(s)", obj.rs()->nSelected());
	obj.rs()->mirrorSelectionLocal(0);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Flip selection's y-coordinates
bool Command::function_FlipY(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	obj.rs()->beginUndoState("Flip y-coordinates of %i atom(s)", obj.rs()->nSelected());
	obj.rs()->mirrorSelectionLocal(1);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Flip selection's z-coordinates
bool Command::function_FlipZ(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	obj.rs()->beginUndoState("Flip z-coordinates of %i atom(s)", obj.rs()->nSelected());
	obj.rs()->mirrorSelectionLocal(2);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Convert coordinates from one reference frame to another
bool Command::function_MatrixConvert(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	// Determine which data has been supplied
	Matrix source, target;
	Vec3<double> o, v;
	bool sourcenoz = FALSE, targetnoz = FALSE;
	int n;
	Atom *i, *j;
	switch (c->nArgs())
	{
		// Twelve atom ids defining both matrices (and optional origin)
		case (12):
		case (15):
			// Determine axes for the matrices
			for (n=0; n<3; n++)
			{
				// Source matrix
				if ((n == 2) && (c->argi(n*2) == 0) && (c->argi(n*2+1) == 0)) sourcenoz = TRUE;
				else
				{
					i = obj.rs()->atom(c->argi(n*2)-1);
					j = obj.rs()->atom(c->argi(n*2+1)-1);
					if ((i == NULL) || (j == NULL)) return FALSE;
					v = obj.rs()->cell()->mimVector(i,j);
					v.normalise();
					source.setColumn(n, v, 0.0);
				}
				// Target matrix
				if ((n == 2) && (c->argi(n*2+6) == 0) && (c->argi(n*2+7) == 0)) targetnoz = TRUE;
				else
				{
					i = obj.rs()->atom(c->argi(n*2+6)-1);
					j = obj.rs()->atom(c->argi(n*2+7)-1);
					if ((i == NULL) || (j == NULL)) return FALSE;
					v = obj.rs()->cell()->mimVector(i,j);
					v.normalise();
					target.setColumn(n, v, 0.0);
				}
				// Adjust Y and Z axes
				if (n == 1)
				{
					source.orthogonaliseColumn(1, 0);
					target.orthogonaliseColumn(1, 0);
				}
				else if (n == 2)
				{
					if (sourcenoz) v = source.columnAsVec3(0) * source.columnAsVec3(1);
					else source.orthogonaliseColumn(2, 0, 1);
					if (targetnoz) v = target.columnAsVec3(0) * target.columnAsVec3(1);
					else target.orthogonaliseColumn(2, 0, 1);
				}
			}
			// Get origin if provided
			if (c->nArgs() == 15) o.set(c->argd(12), c->argd(13), c->argd(14));
			break;
		case (18):
		case (21):
			for (n=0; n<9; n++) source[(n/3)*4 + n%3] = c->argd(n);
			for (n=9; n<18; n++) target[((n-9)/3)*4 + n%3] = c->argd(n);
			// Get origin if provided
			if (c->nArgs() == 21) o.set(c->argd(18), c->argd(19), c->argd(20));
			break;
		default:
			return FALSE;
			break;
	}
	if (msg.isOutputActive(Messenger::Verbose))
	{
		msg.print("Source matrix:\n");
		source.print();
		msg.print("Target matrix:\n");
		target.print();
	}
	// Generate necessary rotation matrix
	source.invert();
	Matrix rotmat = target * source;
	if (msg.isOutputActive(Messenger::Verbose))
	{
		msg.print("Generated rotation matrix:\n");
		rotmat.print();
	}
	// Perform transformation
	obj.rs()->beginUndoState("Transform %i atom(s)", obj.rs()->nSelected());
	obj.rs()->matrixTransformSelection(o, rotmat);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Transform coordinates using supplied matrix / origin
bool Command::function_MatrixTransform(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	// Determine which data has been supplied
	Matrix mat;
	Vec3<double> o;
	int n;
	switch (c->nArgs())
	{
		// Six atom ids defining matrix axes (and optional origin)
		case (9):
		case (12):
			for (n=0; n<9; n++) mat[(n/3)*4+n%3] = c->argd(n);
			// Get origin if provided
			if (c->nArgs() == 12) o.set(c->argd(9), c->argd(10), c->argd(11));
			break;
		default:
			printf("Internal Error: 'matrixtransform' was passed a strange (%i) number of arguments.\n", c->nArgs());
			return FALSE;
	}
	// Perform transformation
	obj.rs()->beginUndoState("Transform %i atom(s)", obj.rs()->nSelected());
	obj.rs()->matrixTransformSelection(o, mat);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Mirror selection along specified axis
bool Command::function_Mirror(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	obj.rs()->beginUndoState("Mirror %i atoms along %c", obj.rs()->nSelected(), 88+c->argi(0));
	obj.rs()->mirrorSelectionLocal(c->argi(0));
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Convert coordinates (specified in atoms/ids) from one reference frame to another
bool Command::function_Reorient(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	// Determine which data has been supplied
	Matrix source, target;
	Vec3<double> o, v;
	bool sourcenoz = FALSE;
	int n;
	Atom *i = NULL, *j = NULL;

	// Determine source matrix
	for (n=0; n<5; n+=2)
	{
		if (c->argType(n) == VTypes::IntegerData)
		{
			if ((n == 4) && (c->argi(n) == 0)) sourcenoz = TRUE;
			else i = obj.rs()->atom(c->argi(n)-1);
		}
		else i = (Atom*) c->argp(n, VTypes::AtomData);
		if (c->argType(n+1) == VTypes::IntegerData)
		{
			if ((n == 4) && (c->argi(n+1) == 0)) sourcenoz = TRUE;
			else j = obj.rs()->atom(c->argi(n+1)-1);
		}
		else j = (Atom*) c->argp(n+1, VTypes::AtomData);
		if ((i == NULL) || (j == NULL)) return FALSE;
		v = obj.rs()->cell()->mimVector(i,j);
		v.normalise();
		source.setColumn(n/2, v, 0.0);
	}
	// Orthogonalise matrix
	source.orthogonaliseColumn(1, 0);
	if (sourcenoz) source.setColumn(2, source.columnAsVec3(0) * source.columnAsVec3(1), 0.0);
	else source.orthogonaliseColumn(2, 0, 1);

	// Determine target matrix
	for (n=0; n<9; ++n) target[(n/3)*4+n%3] = c->argd(n+6);

	// Get origin if provided
	if (c->nArgs() == 18) o.set(c->argd(15), c->argd(16), c->argd(17));

	if (msg.isOutputActive(Messenger::Verbose))
	{
		msg.print("Source matrix:\n");
		source.print();
		msg.print("Target matrix:\n");
		target.print();
	}
	// Generate necessary rotation matrix
	target = target.transpose();
	Matrix rotmat = target * source;
	// Perform transformation
	obj.rs()->beginUndoState("Reorient %i atom(s)", obj.rs()->nSelected());
	obj.rs()->matrixTransformSelection(o, rotmat);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Alter angle between three specified atoms
bool Command::function_SetAngle(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	rv.reset();
	Atom *i = c->argType(0) == VTypes::IntegerData ? obj.rs()->atom(c->argi(0)-1) : (Atom*) c->argp(0, VTypes::AtomData);
	if (i == NULL)
	{
		msg.print("Atom 'i' given to 'setangle' is NULL.\n");
		return FALSE;
	}
	Atom *j = c->argType(1) == VTypes::IntegerData ? obj.rs()->atom(c->argi(1)-1) : (Atom*) c->argp(1, VTypes::AtomData);
	if (j == NULL)
	{
		msg.print("Atom 'j' given to 'setangle' is NULL.\n");
		return FALSE;
	}
	Atom *k = c->argType(2) == VTypes::IntegerData ? obj.rs()->atom(c->argi(2)-1) : (Atom*) c->argp(2, VTypes::AtomData);
	if (k == NULL)
	{
		msg.print("Atom 'k' given to 'setangle' is NULL.\n");
		return FALSE;
	}
	// Clear any current marked selection
	obj.rs()->selectNone(TRUE);
	// Find bond (if any) between i and j
	Bond *b = i->findBond(j);
	// Perform mark-only tree select on atom j, excluding any bond to atom i
	obj.rs()->selectTree(k, TRUE, FALSE, b);
	// If atom 'i' is now marked, there is a cyclic route connecting the two atoms and we can't proceed
	if (i->isSelected(TRUE))
	{
		msg.print("Can't alter the angle of three atoms i-j-k where 'i' and 'k' exist in the same cyclic moiety, or are unbound and within the same fragment.\n");
		return FALSE;
	}
	// Get current angle between the three atoms
	double angle = obj.rs()->angle(i,j,k);
	// Get cross product of bond vectors to define rotation axis
	Vec3<double> v = obj.rs()->cell()->mimVector(j,k) * obj.rs()->cell()->mimVector(j,i);
	v.normalise();
	double delta = c->argd(3) - angle;
	obj.rs()->beginUndoState("Set angle between atoms");
	obj.rs()->rotateSelectionVector(j->r(), v, delta, TRUE);
	obj.rs()->endUndoState();
	return TRUE;
}

// Alter distance between two specified atoms
bool Command::function_SetDistance(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	rv.reset();
	Atom *i = c->argType(0) == VTypes::IntegerData ? obj.rs()->atom(c->argi(0)-1) : (Atom*) c->argp(0, VTypes::AtomData);
	if (i == NULL)
	{
		msg.print("Atom 'i' given to 'setdistance' is NULL.\n");
		return FALSE;
	}
	Atom *j = c->argType(1) == VTypes::IntegerData ? obj.rs()->atom(c->argi(1)-1) : (Atom*) c->argp(1, VTypes::AtomData);
	if (j == NULL)
	{
		msg.print("Atom 'j' given to 'setdistance' is NULL.\n");
		return FALSE;
	}
	// Clear any current marked selection
	obj.rs()->selectNone(TRUE);
	// Find bond (if any) between i and j
	Bond *b = i->findBond(j);
	// Perform mark-only tree select on atom j, excluding any bond to atom i
	obj.rs()->selectTree(j, TRUE, FALSE, b);
	// If atom 'i' is now marked, there is a cyclic route connecting the two atoms and we can't proceed
	if (i->isSelected(TRUE))
	{
		msg.print("Can't alter the distance of two atoms i-j that exist in the same cyclic moiety, or are unbound and within the same fragment.\n");
		return FALSE;
	}
	// Grab the minimum image vector between the two atoms, and shift all those currently marked
	Vec3<double> v = obj.rs()->cell()->mimVector(i,j);
	double delta = c->argd(2) - v.magnitude();
	v.normalise();
	v *= delta;
	obj.rs()->beginUndoState("Set distance between atoms");
	obj.rs()->translateSelectionLocal(v, TRUE);
	obj.rs()->endUndoState();
	return TRUE;
}

// Alter torsion between four specified atoms
bool Command::function_SetTorsion(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	rv.reset();
	Atom *i = c->argType(0) == VTypes::IntegerData ? obj.rs()->atom(c->argi(0)-1) : (Atom*) c->argp(0, VTypes::AtomData);
	if (i == NULL)
	{
		msg.print("Atom 'i' given to 'settorsion' is NULL.\n");
		return FALSE;
	}
	Atom *j = c->argType(1) == VTypes::IntegerData ? obj.rs()->atom(c->argi(1)-1) : (Atom*) c->argp(1, VTypes::AtomData);
	if (j == NULL)
	{
		msg.print("Atom 'j' given to 'settorsion' is NULL.\n");
		return FALSE;
	}
	Atom *k = c->argType(2) == VTypes::IntegerData ? obj.rs()->atom(c->argi(2)-1) : (Atom*) c->argp(2, VTypes::AtomData);
	if (k == NULL)
	{
		msg.print("Atom 'k' given to 'settorsion' is NULL.\n");
		return FALSE;
	}
	Atom *l = c->argType(3) == VTypes::IntegerData ? obj.rs()->atom(c->argi(3)-1) : (Atom*) c->argp(3, VTypes::AtomData);
	if (l == NULL)
	{
		msg.print("Atom 'l' given to 'settorsion' is NULL.\n");
		return FALSE;
	}
	// Clear any current marked selection
	obj.rs()->selectNone(TRUE);
	// Find bond (if any) between j and k
	Bond *b = j->findBond(k);
	// Perform mark-only tree select on atom l, excluding any bond to atom i
	obj.rs()->selectTree(l, TRUE, FALSE, b);
	// If atom 'i' is now marked, there is a cyclic route connecting the two atoms and we can't proceed
	if (i->isSelected(TRUE))
	{
		msg.print("Can't alter the angle of four atoms i-j-k-l where 'i' or 'j' exists in the same cyclic moiety as 'k' or 'l', or are unbound and within the same fragment.\n");
		return FALSE;
	}
	// Get current torsion between the four atoms
	double angle = obj.rs()->torsion(i,j,k,l);
	// Rotation vector will be vector k->j
	Vec3<double> v = obj.rs()->cell()->mimVector(k,j);
	v.normalise();
	double delta = c->argd(4) - angle;
	obj.rs()->beginUndoState("Set torsion between atoms");
	obj.rs()->rotateSelectionVector(j->r(), v, delta, TRUE);
	obj.rs()->endUndoState();
	return TRUE;
}

// Translate current selection in local coordinates ('translate dx dy dz')
bool Command::function_Translate(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	Vec3<double> tvec = c->arg3d(0);
	obj.rs()->beginUndoState("Translate Cartesian (%i atom(s), %f %f %f)", obj.rs()->nSelected(), tvec.x, tvec.y, tvec.z);
	obj.rs()->translateSelectionLocal(tvec);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Translate activeatom ('translateatom <dx dy dz>')
bool Command::function_TranslateAtom(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::AtomPointer)) return FALSE;
	Vec3<double> tvec = c->arg3d(0);
	obj.rs()->beginUndoState("Translate Cartesian (atom %i, %f %f %f)", obj.i->id()+1, tvec.x, tvec.y, tvec.z);
	obj.rs()->translateAtom(obj.i, tvec);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Translate current selection in fractional cell coordinates ('translatecell dx dy dz')
bool Command::function_TranslateCell(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	Vec3<double> tvec;
	tvec = obj.rs()->cell()->axes() * c->arg3d(0);
	obj.rs()->beginUndoState("Translate Cell (%i atom(s), %f %f %f)", obj.rs()->nSelected(), tvec.x, tvec.y, tvec.z);
	obj.rs()->translateSelectionLocal(tvec);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}

// Translate current selection in world coordinates ('translateworld dx dy dz')
bool Command::function_TranslateWorld(CommandNode *c, Bundle &obj, ReturnValue &rv)
{
	if (obj.notifyNull(Bundle::ModelPointer)) return FALSE;
	Vec3<double> tvec;
	tvec = c->arg3d(0);
	obj.rs()->beginUndoState("Translate World (%i atom(s), %f %f %f)", obj.rs()->nSelected(), tvec.x, tvec.y, tvec.z);
	obj.rs()->translateSelectionWorld(tvec);
	obj.rs()->endUndoState();
	rv.reset();
	return TRUE;
}
